<?php

namespace Drupal\term_glossary\Service;

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\Unicode;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigManagerInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\Entity\EntityViewDisplay;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FormatterInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\taxonomy\TermInterface;

/**
 * The TermGlossary manager service.
 */
class TermGlossaryManager implements TermGlossaryManagerInterface {

  use StringTranslationTrait;

  const MATCH_DEFAULT = -1;
  const MATCH_DISABLE = 0;
  const MATCH_ENFORCE = 1;

  /**
   * The config manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected ConfigManagerInterface $configManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The entity term storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $termStorage;

  /**
   * The cache service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected CacheBackendInterface $cache;

  /**
   * The configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * The language manager service.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected LanguageManagerInterface $languageManager;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * The cache id.
   *
   * @var string
   */
  protected string $cacheId = 'glossary_terms_array';

  /**
   * The seen terms.
   *
   * @var array
   */
  protected static array $seenTerms = [];

  public function __construct(ConfigManagerInterface $configManager, EntityTypeManagerInterface $entityTypeManager, CacheBackendInterface $cache, LanguageManagerInterface $language_manager, LoggerChannelInterface $logger) {
    $this->configManager = $configManager;
    $this->entityTypeManager = $entityTypeManager;
    $this->cache = $cache;
    $this->termStorage = $this->entityTypeManager->getStorage('taxonomy_term');
    $this->config = $configManager->getConfigFactory()->get('term_glossary.glossaryconfig');
    $this->languageManager = $language_manager;
    $this->logger = $logger;
  }

  /**
   * Builds the regexp pattern to match a term.
   */
  protected function buildTermPattern(array $term_array, bool $global_full_word, bool $global_case_sensitive) {
    $term_full_word = $term_array['full_word'] ?? self::MATCH_DEFAULT;
    $term_case_sensitive = $term_array['case_sensitive'] ?? self::MATCH_DEFAULT;

    $full_word = ($term_full_word === self::MATCH_ENFORCE)
      || ($global_full_word && ($term_full_word !== self::MATCH_DISABLE));
    $case_sensitive = ($term_case_sensitive === self::MATCH_ENFORCE)
      || ($global_case_sensitive && ($term_case_sensitive !== self::MATCH_DISABLE));

    $name = preg_quote($term_array['name'], '/');
    $pattern = $full_word ? "/\b($name)\b/u" : "/($name)/u";
    if (!$case_sensitive) {
      $pattern .= 'i';
    }

    return $pattern;
  }

  /**
   * {@inheritDoc}
   */
  public function replaceFieldValue($input, $vocabularies, $root_entity = NULL) {
    if (empty($input)) {
      return $input;
    }
    else {
      if (!Unicode::validateUtf8($input)) {
        $this->logger->debug($this->t('Input is not valid UTF-8: @input', ['@input' => $input]));
        return $input;
      }

      $vocabularies = is_array($vocabularies) ? $vocabularies : [$vocabularies];

      $term_list = [];
      foreach ($vocabularies as $vocabulary) {
        $term_list += $this->getTerms($vocabulary);
      }

      $langcode = isset($root_entity)
        ? $root_entity->language()->getId()
        : $this->languageManager->getCurrentLanguage()->getId();

      $term_list = array_filter($term_list, function ($term) use ($langcode) {
        return $term['lang'] === $langcode;
      });

      $is_single_match_per_field = $this->config->get('single_match') ?? FALSE;
      $is_single_match_per_content = $this->config->get('single_match_per_content') ?? FALSE;
      $is_full_word = $this->config->get('full_word') ?? FALSE;
      $is_case_sensitive = $this->config->get('case_sensitive') ?? FALSE;

      $seen_id = 0;
      static::$seenTerms[0] = [];

      if ($is_single_match_per_content && isset($root_entity)) {
        $seen_id = $root_entity->id();
        if (!isset(static::$seenTerms[$seen_id])) {
          static::$seenTerms[$seen_id] = [];
        }
      }

      $html_dom = Html::load($input);
      $xpath = new \DOMXPath($html_dom);

      /*
       * Parts and ideas of the code have been taken from the "glossify" module.
       * https://git.drupalcode.org/project/glossify
       */
      $query = '//text()[not(ancestor::a) and not(ancestor::img) and not(ancestor::span[@class="glossary-exclude"]) and not(ancestor::p[@class="glossary-exclude"])]';
      /** @var \DOMNodeList $text_nodes */
      $text_nodes = $xpath->query($query);
      foreach ($text_nodes as $original_node) {
        $text = $original_node->nodeValue;

        if (empty(trim($text))) {
          continue;
        }

        $new_value = $text;
        $parent_node = $original_node->parentNode;
        $next_node = $original_node->nextSibling;
        $parent_node->removeChild($original_node);

        $replacements = [];
        $replacement_index = 0;
        $single_match = $is_single_match_per_field || $is_single_match_per_content;

        foreach ($term_list as $term_id => $term_array) {
          $seen_name = mb_strtolower($term_array['name']);
          if ($single_match && in_array($seen_name, static::$seenTerms[$seen_id])) {
            // We have already replaced that term, skip.
            continue;
          }
          $pattern = $this->buildTermPattern($term_array, $is_full_word, $is_case_sensitive);
          $new_value = preg_replace_callback(
            $pattern,
            function ($matches) use (&$replacements, &$replacement_index, $term_id, $single_match, $seen_id, $seen_name) {
              $replacements[$replacement_index] = $this->buildGlossaryButton($term_id, $matches[1]);
              $replacement = '{{' . $replacement_index . '}}';
              $replacement_index += 1;
              if ($single_match) {
                static::$seenTerms[$seen_id][] = $seen_name;
              }
              return $replacement;
            },
            $new_value,
            $single_match ? 1 : -1,
          );
        }

        // $originalNode->nodeValue unescaped entities, this escapes them back.
        $new_value = Html::escape($new_value);

        $new_value = preg_replace_callback(
          '/\{\{(\d+)}}/',
          function ($matches) use ($replacements) {
            $index = (int) $matches[1];
            return $replacements[$index] ?? '{{' . $index . '}}';
          },
          $new_value
        );

        $frag = $html_dom->createDocumentFragment();
        $frag->appendXML($new_value);

        $parent_node->insertBefore($frag, $next_node);
      }

      return Html::serialize($html_dom);
    }
  }

  /**
   * Build the button for the matched term.
   */
  protected function buildGlossaryButton(int|string $term_id, mixed $match) {
    return '<span role="button" tabindex="0" class="glos-term" data-gterm="' . $term_id . '">' . Html::escape($match) . '</span>';
  }

  /**
   * {@inheritDoc}
   */
  public function getConfigValue(string $key): ?string {
    return $this->config->get($key);
  }

  /**
   * {@inheritDoc}
   */
  public function invalidateTermsCache($vocab) {
    $cacheId = $this->cacheId . '-' . $vocab;
    $this->cache->invalidate($cacheId);
  }

  /**
   * Get an updated list of glossary terms.
   *
   * @return array
   *   Glossary term list
   */
  private function getTerms($vocab) {
    $cacheId = $this->cacheId . '-' . $vocab;
    $data = $this->cache->get($cacheId);
    if (empty($data)) {
      $terms = $this->updateTermList($vocab);
    }
    else {
      $query = $this->termStorage->getQuery();
      $query->condition('vid', $vocab);
      $query->condition('status', 1);
      $query->accessCheck();
      $entity_ids = $query->execute();
      // Someone could have added new term since last cache.
      if (count($entity_ids) === count($data->data)) {
        // Get the list from cache.
        $terms = $data->data;
      }
      else {
        // Set cache again.
        $terms = $this->updateTermList($vocab);
      }

    }
    return $terms;
  }

  /**
   * Update term list.
   *
   * @param string $vocab
   *   Glossary vocabulary id.
   *
   * @return array
   *   Glossary term list
   */
  private function updateTermList($vocab) {
    $terms = $this->termStorage->loadByProperties(['vid' => $vocab, 'status' => 1]);
    // Sort terms by weight to allow processing prioritization.
    usort($terms, function (TermInterface $a, TermInterface $b) {
      return $a->getWeight() <=> $b->getWeight();
    });
    $per_term_options = $this->getConfigValue('per_term_options') ?? FALSE;
    $terms_array = [];
    /**
     * @var \Drupal\taxonomy\TermInterface $term
     */
    foreach ($terms as $term) {
      $term_data = [
        'tid' => $term->id(),
        'name' => $term->getName(),
        'lang' => $term->language()->getId(),
        'full_word' => self::MATCH_DEFAULT,
        'case_sensitive' => self::MATCH_DEFAULT,
      ];
      if ($per_term_options) {
        if ($term->hasField('field_full_word')) {
          if (!$term->get('field_full_word')->isEmpty()) {
            $term_data['full_word'] = (int) $term->get('field_full_word')->value;
          }
        }
        if ($term->hasField('field_case_sensitive')) {
          if (!$term->get('field_case_sensitive')->isEmpty()) {
            $term_data['case_sensitive'] = (int) $term->get('field_case_sensitive')->value;
          }
        }
      }
      $terms_array[$term->id()] = $term_data;
    }
    $cacheId = $this->cacheId . '-' . $vocab;
    $this->cache->set($cacheId, $terms_array, CacheBackendInterface::CACHE_PERMANENT);
    return $terms_array;
  }

  /**
   * {@inheritDoc}
   */
  public function getVocabulariesFromFieldPreprocessVariables(&$variables) {
    $entity = $variables['element']['#object'];
    $view_mode = $variables['element']['#view_mode'];
    $field_name = $variables['element']['#field_name'];
    $display_vocabularies = $this->getVocabulariesFromEntity($entity, $field_name, $view_mode);
    if ($display_vocabularies === FALSE) {
      return FALSE;
    }
    else {
      if (empty($display_vocabularies)) {
        return explode(',', $this->config->get('vocab'));
      }
      else {
        return $display_vocabularies;
      }
    }
  }

  /**
   * Get the vocabulary from the field display.
   *
   * @param array|null $field_display
   *   The field display to get the vocabulary from.
   *
   * @return array|false
   *   The vocabulary name.
   */
  protected function getVocabulariesFromFieldDisplay(?array $field_display) {
    if (isset($field_display['third_party_settings']['term_glossary'])) {
      if ($field_display['third_party_settings']['term_glossary']['glossary']) {
        if (empty($field_display['third_party_settings']['term_glossary']['glossary_vocabulary'])) {
          return [];
        }
        else {
          $vocabularies_ids = [];
          $glossary_vocabularies = $field_display['third_party_settings']['term_glossary']['glossary_vocabulary'];
          if (is_array($glossary_vocabularies)) {
            foreach ($glossary_vocabularies as $vocabulary) {
              $vocabularies_ids[] = $vocabulary['target_id'];
            }
          }
          else {
            $vocabularies_ids[] = (string) $glossary_vocabularies;
          }
          return $vocabularies_ids;
        }
      }
    }
    return FALSE;
  }

  /**
   * Find the vocabulary associated with the display field of an entity.
   *
   * @param mixed $entity
   *   The entity.
   * @param string $field_name
   *   The field name.
   * @param string $view_mode
   *   The view mode.
   *
   * @return false|array
   *   The vocabulary if defined, NULL if not defined, FALSE if not active.
   */
  protected function getVocabulariesFromEntity($entity, $field_name, $view_mode) {
    // Get the field formatter settings.
    $entity_display = EntityViewDisplay::collectRenderDisplay($entity, $view_mode);
    $field_display = $entity_display->getComponent($field_name);
    return $this->getVocabulariesFromFieldDisplay($field_display);
  }

  /**
   * {@inheritDoc}
   */
  public function getFieldFormatterThirdPartySettingsElements(
    FormatterInterface $plugin,
    FieldDefinitionInterface $field_definition,
  ) {
    $elements = [];
    $elements['glossary'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable term glossary'),
      '#default_value' => $plugin
        ->getThirdPartySetting('term_glossary', 'glossary'),
      '#description' => $this->t('Enable the term glossary functionality for this field.'),
      '#attributes' => [
        'class' => ['toggle-vocabulary-autocomplete'],
        'data-toggle-target' => $field_definition->getName() . '_glossary_vocabulary',
      ],
    ];
    $configuration_url = Url::fromRoute('term_glossary.glossary_config_form')->toString();
    $glossary_vocabularies = $plugin->getThirdPartySetting('term_glossary', 'glossary_vocabulary');
    $vocabularies = $this->getVocabulariesFromThirdPartySetting($glossary_vocabularies);
    $elements['glossary_vocabulary'] = [
      '#type' => 'entity_autocomplete',
      '#target_type' => 'taxonomy_vocabulary',
      '#default_value' => $vocabularies,
      '#description' => $this->t('Select the vocabularies to use as a glossary for this field content.<br/>If none are set, the global setting applies (<a href=":url">Settings page</a>).', [':url' => $configuration_url]),
      '#tags' => TRUE,
      '#attributes' => [
        'data-toggle-target-id' => $field_definition->getName() . '_glossary_vocabulary',
      ],
    ];
    $elements['#attached']['library'][] = 'term_glossary/glossary.toggle_autocomplete';
    return $elements;
  }

  /**
   * Gets the vocabularies referenced in the third-party setting.
   */
  private function getVocabulariesFromThirdPartySetting(mixed $glossary_vocabularies) {
    if (empty($glossary_vocabularies)) {
      $vocabularies = NULL;
    }
    else {
      $vocabularies_ids = [];
      if (is_array($glossary_vocabularies)) {
        foreach ($glossary_vocabularies as $vocabulary) {
          $vocabularies_ids[] = $vocabulary['target_id'];
        }
      }
      else {
        $vocabularies_ids[] = (string) $glossary_vocabularies;
      }
      $vocabularies = Vocabulary::loadMultiple($vocabularies_ids);
    }
    return $vocabularies;
  }

  /**
   * Gets the root entity of a paragraph.
   *
   * @param mixed $entity
   *   The current entity to check for parents.
   *
   * @return mixed
   *   The root entity.
   */
  protected function findRootEntity($entity) {
    if ($entity->getEntityTypeId() === 'paragraph') {
      $parent_entity = $entity->getParentEntity();
      $root = $this->findRootEntity($parent_entity);
    }
    else {
      $root = $entity;
    }
    return $root;
  }

  /**
   * {@inheritDoc}
   */
  public function getRootEntityFromFieldPreprocessVariables($variables) {
    if (isset($variables['element']['#object'])) {
      return $this->findRootEntity($variables['element']['#object']);
    }
    return NULL;
  }

}
